"""This module contains the objects to handle the fractal tables V1."""

from ngio.tables.v1.feature_tables import FeatureTableV1
from ngio.tables.v1.masking_roi_tables import MaskingROITableV1
from ngio.tables.v1.roi_tables import ROITableV1

__all__ = ["ROITableV1", "FeatureTableV1", "MaskingROITableV1"]
